with Ada.Text_IO;
use  Ada.Text_IO;

package body Fahrkartenautomat is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   1.0
  -- Date      12 April 2014
  --====================================================================
  -- Demonstration Trennung der Verantwortlichkeiten, insbesondere
  -- der Ablaufsteuerung vom Rest des Kodes.
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  12.04.2014 Created for Ada tasking course
  --====================================================================

  package Geld_Arithmetik is
    subtype Automatenspeicher is Stueckelung (Akzeptiertes_Geldstueck);
    function "+" (L, R: Automatenspeicher) return Automatenspeicher;
    function "-" (L, R: Automatenspeicher) return Automatenspeicher;
  end Geld_Arithmetik;

  package body Geld_Arithmetik is separate;

  use Geld_Arithmetik;

  --------------------------------------------------------------
  -- Die Maschine steuert nur den Ablauf, Berechnungen werden
  -- auerhalb gemacht.

  task Maschine is
    entry Fahrkarte_gewaehlt;
    entry Geldeingabe;
    entry Abbruch;
    entry Ausgabe;
  end Maschine;

  task body Maschine is
  begin
    Lebenslang:
    loop
      accept Fahrkarte_gewaehlt;
      Geld_einsammeln:
      loop
        select  -- Kundeneingabe oder Zeitberschreitung
          accept Geldeingabe;
          select  -- Geldstck prfen (dauert ein wenig)
            accept Abbruch;  -- genug eingeworfen
            exit Geld_einsammeln;
          or
            delay 0.1;  -- weitere Eingabe erforderlich
          end select;
        or
          accept Abbruch;  -- durch Kunden
          exit Geld_einsammeln;
        or
          delay 15.0;  -- zu lange keine Eingabe
          exit Geld_einsammeln;
        end select;
      end loop Geld_einsammeln;
      accept Ausgabe;
    end loop Lebenslang;
  end Maschine;

  --------------------------------------------------------------

  function Image (Preis: Natural) return String is
    Zwerg: constant String := Integer'Image (Preis);
  begin
    if Preis = 0 then
      return "0";
    else     -- Vorkomma
      return (if Zwerg'Last > 3 then Zwerg (1 .. Zwerg'Last - 2) else " 0") &
             -- Nachkomma
             '_' & Zwerg (Zwerg'Last - 1 .. Zwerg'Last);
    end if;
  end Image;

  -- Nach der Fahrkartenwahl:
  --   Der Mnzschlitz ffnet sich; Preis wird angezeigt.
  -- Geldeingabe abwarten.
  --   Prfung des Geldstcks auf Echtheit. (Wie wird Falschgeld simuliert?)
  --   Prfung der eingegebenen Summe.
  --   Wenn genug:
  --     Mnzschlitz schliet sich;
  --     Fahrkarten- und Wechselgeldausgabe.
  --   Sonst Restpreis anzeigen.
  -- Wenn statt Geld Abbruch gewhlt wird:
  --   Mnzschlitz schliet sich;
  --   Ausgabe der eingegebenen Geldstcke.
  -- Wenn lngere Zeit keine Eingabe erfolgt:
  --   Mnzschlitz schliet sich;
  --   Automatischer Abbruch und Geldrckgabe.

  --  Was soll passieren, wenn nicht genug Wechselgeld im Automaten ist?

  Fahrkartenpreis: constant array (Fahrkarte) of Positive :=
    (Kurzstrecke  =>  1_20,
     Innenraum    =>  2_50,
     Gesamtgebiet =>  4_90,
     Tageskarte   => 10_20);

  Speicher: Automatenspeicher :=
    (Cent_10  => 200,
     Cent_20  => 100,
     Cent_50  =>  50,
     Euro_1   => 100,
     Euro_2   =>  50,
     Euro_5   =>  50,
     Euro_10  =>   0);
--  Speicherwert: Positive := Betrag (Speicher);

  -- Kundendaten
  Wahl : Fahrkartenoption := Keine;
  Preis: Positive;
  Kunde: Automatenspeicher := (others => 0);
  Summe: Natural           := Betrag (Kunde);

  procedure Waehle (F: Fahrkarte) is
  begin
    if Wahl = Keine then
      Wahl := F;
      Preis := Fahrkartenpreis (Wahl);
      Put_Line (Fahrkarte'Image (Wahl) & ": Preis " & Image (Preis));
      Maschine.Fahrkarte_gewaehlt;
    else
      Put_Line ("Taste auer Funktion");
    end if;
  end Waehle;

  procedure Geldeingabe (S: Akzeptiertes_Geldstueck; Ergebnis: out Boolean) is
  begin
    if Wahl = Keine then
      Put_Line ("Geldeingabeschlitz geschlossen");
      Ergebnis := False;
      return;
    end if;
    Ergebnis := True;
    Maschine.Geldeingabe;
    Kunde (S) := Kunde (S) + 1;
    Summe     := Summe + Wert (S);
    if Summe > Preis then
      Maschine.Abbruch;
    else
      Put_Line ("Noch zu bezahlen:" & Image (Preis - Summe));
    end if;
  end Geldeingabe;

  procedure Abbruch is
  begin
    if Wahl = Keine then
      Put_Line ("Taste auer Funktion");
    else
      Maschine.Abbruch;
    end if;
  end Abbruch;

  procedure Ausgabe (Wechselgeld   : out Stueckelung;
                     eine_Fahrkarte: out Fahrkartenoption) is
  begin
    Maschine.Ausgabe;
    if Preis > Summe then  -- Abbruch
      Wechselgeld    := Kunde;
      eine_Fahrkarte := Keine;
    else
      Speicher := Speicher + Kunde;
      eine_Fahrkarte := Wahl;
      Wechselgeld  := Stueckle (Summe - Preis) (Cent_10 .. Euro_10);
      Speicher     := Speicher - Wechselgeld;
    end if;
    -- Auf Anfang setzen
    Wahl  := Keine;
    Summe := 0;
    Kunde := (others => 0);
  end Ausgabe;

end Fahrkartenautomat;
