------------------------------------------------------------------------
-- This software is provided as is in the hope that it might be found
-- useful.
-- You may use and modify it freely provided you keep this copyright
-- notice unchanged and mark modifications appropriately.
--
-- Bug reports and proposals for improvements are welcome. Please send
-- them to the eMail address below.
--
-- Christoph Karl Walter Grein
-- Hauptstr. 42
-- D-86926 Greifenberg
-- Germany
--
-- eMail: christ-usch.grein@t-online.de
--
-- Copyright (c) 2023 Christoph Karl Walter Grein
------------------------------------------------------------------------

with Ada.Unchecked_Deallocation;
with Ada.Text_IO;

procedure Task_Deallocation is

  --====================================================================
  -- Autor     Christoph Grein
  -- Version   1.0
  -- Datum     8 April 2023
  --====================================================================
  -- Die Aussagen des Ada-Referenzmanual dazu sind verstreut. Hier sind
  -- die relevanten Stellen zusammengefasst.
  --
  -- 13.11.2(9/5) "3. Free(X), when X is not equal to null first per-
  -- forms finalization of the object designated by X ..., as described
  -- in 7.6.1. It then deallocates the storage occupied by the object
  -- designated by X. ... There is one exception: if the object being
  -- freed contains tasks, it is unspecified whether the object is
  -- deallocated."
  --
  -- 13.11.2(17) "For a standard storage pool, Free should actually
  -- reclaim the storage."
  -- GNAT Reference Manual 6.31: Followed
  --
  -- 7.6.1(1, 3/5) A task_type is a 'master construct'. Its execution
  -- is a 'master'.
  -- "A master is finalized after it is complete, and before it is
  -- left."
  --
  -- 7.6.1(4) "For the 'finalization' of a master, dependent tasks are
  -- first awaited, as explained in 9.3."
  --
  -- 3.2(4/2) A task type is a 'composite type'.
  --
  -- 7.6.1(5, 9/3) "For the 'finalization' of an object:"
  -- "If the full type of the object is a composite type, then after
  -- performing the above actions, if any, every component of the
  -- object is finalized in an arbitrary order, except as follows:
  -- if the object has a component with an access discriminant con-
  -- strained by a per-object expression, this component is finalized
  -- before any components that do not have such discriminants; for an
  -- object with several components with such a discriminant, they are
  -- finalized in the reverse of the order of their component_declarat-
  -- ions;"
  --
  -- 7.6.1(10) "Immediately before an instance of Unchecked_Dealloca-
  -- tion reclaims the storage of an object, the object is finalized."
  --
  -- 9.3(2/4) The master of R is the master that includes the elabora-
  -- tion of the declaration of the access type.
  --
  -- 9.3(5) "A task is said to be completed when the execution of its
  -- corresponding task_body is completed. A task is said to be termi-
  -- nated when any finalization of the task_body has been performed
  -- (see 7.6.1). The first step of finalizing a master (including a
  -- task_body) is to wait for the termination of any tasks dependent
  -- on the master. The task executing the master is blocked until all
  -- the dependents have terminated. Any remaining finalization is then
  -- performed and the master is left."
  --
  -- Fazit:
  -- * Prozesse laufen trotz UD weiter bis zum Ende.
  -- * Sie sind nicht mehr referenzierbar.
  -- * Ob der Speicherplatz freigegeben wird, ist unspezifiziert im
  --   Ada-Referenzmanual.
  -- * Deallokieren Sie also erst, wenn der Prozess terminiert ist.
  --====================================================================
  -- Geschichte:
  -- Autor  Version  Datum    Grund der nderung
  --  C.G.   1.0  08.04.2023  Erzeugt fr Ada-Tasking-Kurs
  --====================================================================

  task type Run;
  task body Run is
  begin
    for I in 1 .. 20 loop
      Ada.Text_IO.Put_Line ("Run" & I'Image);
      delay 1.0;
    end loop;
  end Run;

  type Run_Access is access Run;
  procedure Free is new Ada.Unchecked_Deallocation (Run, Run_Access);

  G: Run_Access;

begin

  declare  -- nicht der Meister

    R: Run_Access := new Run;  -- Der Meister ist die Prozedur;
                               -- dort ist der Zugriffstyp definiert.
  begin

    G := R;  -- G zeigt auf dasselbe Objekt wie R

    delay 10.0;
    Ada.Text_IO.Put_Line ("Free (R)");  -- Der Prozess luft weiter
    Free (R);                           -- bis zum Ende, obwohl er
    Ada.Text_IO.Put_Line ("freed");     -- deallokiert ist.

    -- R.all = 0.
    -- Die Dereferenzierung mit G.all ist jetzt illegal.

  end;  -- kann sofort verlassen werden

  Ada.Text_IO.Put_Line ("terminated?");

end Task_Deallocation;
