------------------------------------------------------------------------
-- This software is provided as is in the hope that it might be found
-- useful.
-- You may use and modify it freely provided you keep this copyright
-- notice unchanged and mark modifications appropriately.
--
-- Bug reports and proposals for improvements are welcome. Please send
-- them to the eMail address below.
--
-- Christoph Karl Walter Grein
-- Hauptstr. 42
-- D-86926 Greifenberg
-- Germany
--
-- eMail: Christ-Usch.Grein@T-Online.de
--
-- Copyright (c) 2009, 2014, 2023 Christoph Karl Walter Grein
------------------------------------------------------------------------

with Ada.Text_IO;
use  Ada.Text_IO;

procedure Timed_Entry_Call is

  --====================================================================
  -- Author    Christoph Grein
  -- Version   1.0
  -- Date      8 May 2014
  --====================================================================
  -- This program simulates case B.
  --
  -- The times at (1) and (2) are just for the ease of experimenting;
  -- in real code, this would most probably be microseconds.
  -- (1) is in effect a race condition and the rendezvous might just
  -- start a tiny fraction before the time-out occurs (it scrapes
  -- through, so to speak), so that determining to whom to requeue takes
  -- longer than the time available before the time-out.
  --
  -- Expected result:
  --
  -- Round 1
  -- E1 accepted
  -- E1 requeued
  -- First time aborting because E2 not ready
  -- E2 ready
  --
  -- Round 2
  -- E1 accepted
  -- E1 requeued
  -- E2 accepted
  -- E1 done
  --
  -- Done
  --====================================================================
  -- History
  -- Author Version   Date    Reason for change
  --  C.G.    1.0  08.05.2014 Reworked for Ada tasking course
  --  C.G.    0.0  13.01.2009 Created for experimenting
  --====================================================================

  task Server is
    entry E1;
  end Server;

  task Hidden is
    entry Start;
    entry E2;
  end Hidden;

  task body Server is
  begin
    for I in 1 .. 2 loop
      accept E1 do
        Put_Line ("E1 accepted");
        delay 3.0;  -- (1) do some work that takes longer than client's patience
        Put_Line ("E1 requeued");
        requeue Hidden.E2 with abort;
      end E1;
    end loop;
  end Server;

  task body Hidden is
  begin
    accept Start;
    Put_Line ("E2 ready");
    accept E2 do
      Put_Line ("E2 accepted");
    end E2;
  end Hidden;

begin

  for I in 1 .. 2 loop

    Put_Line ("Round" & Integer'Image (I));

    select
      Server.E1;  -- E1 is ready for rendezvous
      Put_Line ("E1 done");
    or
      delay 1.0;  -- (2)
      Put_Line ("First time aborting because E2 not ready");
      Hidden.Start;  -- Make E2 ready for second round
    end select;

    delay 1.0;  -- Wait a bit before starting the second round

    New_Line;

  end loop;

  Put_Line ("Done");

end Timed_Entry_Call;
